import datetime
import typing

import itsdangerous
from starlette.authentication import AuthCredentials, AuthenticationBackend, BaseUser
from starlette.requests import HTTPConnection
from starlette.responses import Response

from starlette_auth.backends import ByIdUserFinder
from starlette_auth.login import get_scopes, LoginScopes

REMEMBER_COOKIE_NAME = "remember_me"

_RT = typing.TypeVar("_RT", bound=Response)


class RememberMeBackend(AuthenticationBackend):
    """Authenticates user using "remember me" cookie.

    It also adds "login:remembered" scope to distinguish between fresh and remembered logins."""

    def __init__(
        self,
        user_loader: ByIdUserFinder,
        secret_key: str,
        duration: datetime.timedelta | None = None,
        *,
        cookie_name: str = REMEMBER_COOKIE_NAME,
    ) -> None:
        self.secret_key = secret_key
        self.user_loader = user_loader
        self.cookie_name = cookie_name
        self.duration = duration

    async def authenticate(self, conn: HTTPConnection) -> tuple[AuthCredentials, BaseUser] | None:
        if cookie_value := conn.cookies.get(self.cookie_name):
            try:
                max_age = int(self.duration.total_seconds()) if self.duration else None
                signer = itsdangerous.TimestampSigner(secret_key=self.secret_key)
                user_id = signer.unsign(cookie_value, max_age=max_age).decode("utf8")
                if user := await self.user_loader(conn, user_id):
                    return AuthCredentials(scopes=get_scopes(user) + [LoginScopes.REMEMBERED]), user
            except itsdangerous.BadSignature:
                return None
        return None


def remember_me(
    response: _RT,
    secret_key: str,
    user: BaseUser,
    duration: datetime.timedelta,
    *,
    cookie_name: str = REMEMBER_COOKIE_NAME,
    cookie_path: str = "/",
    cookie_domain: str | None = None,
    cookie_samesite: typing.Literal["lax", "strict", "none"] = "lax",
    cookie_secure: bool = False,
    cookie_http_only: bool = True,
) -> _RT:
    """Remember user by setting a cookie."""
    signer = itsdangerous.TimestampSigner(secret_key)
    value = signer.sign(user.identity).decode("utf8")
    response.set_cookie(
        key=cookie_name,
        value=value,
        max_age=int(duration.total_seconds()),
        path=cookie_path,
        domain=cookie_domain,
        secure=cookie_secure,
        httponly=cookie_http_only,
        samesite=cookie_samesite,
    )
    return response


def forget_me(
    response: _RT,
    *,
    cookie_name: str = REMEMBER_COOKIE_NAME,
    cookie_path: str = "/",
    cookie_domain: str | None = None,
    cookie_samesite: typing.Literal["lax", "strict", "none"] = "lax",
    cookie_secure: bool = False,
    cookie_http_only: bool = True,
) -> _RT:
    """Forget user by removing 'remember me' cookie."""
    response.set_cookie(
        key=cookie_name,
        value="null",
        max_age=-1,
        path=cookie_path,
        domain=cookie_domain,
        secure=cookie_secure,
        httponly=cookie_http_only,
        samesite=cookie_samesite,
    )
    return response



from starlette_auth.backends import MultiBackend, SessionBackend
from starlette_auth.login import is_authenticated, login, logout
from starlette_auth.middleware import LoginRequiredMiddleware
from starlette_auth.remember_me import forget_me, remember_me, RememberMeBackend
from starlette_auth.scopes import confirm_login, is_confirmed, LoginScopes

__all__ = [
    "login",
    "logout",
    "is_authenticated",
    "MultiBackend",
    "SessionBackend",
    "RememberMeBackend",
    "remember_me",
    "forget_me",
    "confirm_login",
    "is_confirmed",
    "LoginScopes",
    "LoginRequiredMiddleware",
]



from starlette.authentication import AuthCredentials, BaseUser, UnauthenticatedUser
from starlette.requests import HTTPConnection

from starlette_auth.scopes import get_scopes, LoginScopes
from starlette_auth.settings import SESSION_KEY


def _regenerate_session_id(connection: HTTPConnection) -> None:
    if "session_handler" in connection.scope:  # when starsessions installed
        from starsessions import regenerate_session_id

        regenerate_session_id(connection)


async def login(connection: HTTPConnection, user: BaseUser) -> None:
    """Login user."""

    # Regenerate session id to prevent session fixation.
    # Attack vector:
    # A malicious user can steal session ID from victim's browser and set it into his own.
    # When victim sign in and session is NOT regenerated then two browsers will share same session and data
    _regenerate_session_id(connection)

    connection.scope["auth"] = AuthCredentials(scopes=get_scopes(user) + [LoginScopes.FRESH])
    connection.scope["user"] = user
    connection.session[SESSION_KEY] = user.identity


async def logout(connection: HTTPConnection) -> None:
    connection.session.clear()  # wipe all data
    _regenerate_session_id(connection)
    connection.scope["auth"] = AuthCredentials()
    connection.scope["user"] = UnauthenticatedUser()


def is_authenticated(connection: HTTPConnection) -> bool:
    """Check if user is authenticated."""
    value: bool = connection.auth and connection.user.is_authenticated
    return value



import typing

from starlette.authentication import AuthCredentials, AuthenticationBackend, BaseUser
from starlette.requests import HTTPConnection

from starlette_auth.login import get_scopes, SESSION_KEY

ByIdUserFinder = typing.Callable[[HTTPConnection, str], typing.Awaitable[BaseUser | None]]


class SessionBackend(AuthenticationBackend):
    """Authentication backend that uses session to store user information."""

    def __init__(self, user_loader: ByIdUserFinder) -> None:
        self.user_loader = user_loader

    async def authenticate(self, conn: HTTPConnection) -> tuple[AuthCredentials, BaseUser] | None:
        user_id: str = conn.session.get(SESSION_KEY, "")
        if user_id and (user := await self.user_loader(conn, user_id)):
            return AuthCredentials(scopes=get_scopes(user)), user
        return None


class MultiBackend(AuthenticationBackend):
    """Authenticate user using multiple backends."""

    def __init__(self, backends: list[AuthenticationBackend]) -> None:
        self.backends = backends

    async def authenticate(self, conn: HTTPConnection) -> tuple[AuthCredentials, BaseUser] | None:
        for backend in self.backends:
            if result := await backend.authenticate(conn):
                return result
        return None



import enum
import typing

from starlette.authentication import AuthCredentials, BaseUser
from starlette.requests import HTTPConnection

from starlette_auth.settings import SESSION_KEY


class UserWithScopes(typing.Protocol):  # pragma: no cover
    def get_scopes(self) -> list[str]: ...


def get_scopes(user: BaseUser | UserWithScopes) -> list[str]:
    """Extract scopes from user object."""
    if hasattr(user, "get_scopes"):
        return user.get_scopes()
    return []


class LoginScopes(enum.StrEnum):
    FRESH = "login:fresh"
    REMEMBERED = "login:remembered"


def confirm_login(connection: HTTPConnection) -> None:
    """Convert remembered login to fresh login.
    Fresh login is the one where user provided credentials."""
    credentials: AuthCredentials = connection.auth
    if LoginScopes.REMEMBERED in credentials.scopes:
        credentials.scopes.remove(LoginScopes.REMEMBERED)
        credentials.scopes.append(LoginScopes.FRESH)
        connection.session[SESSION_KEY] = connection.user.identity


def is_confirmed(connection: HTTPConnection) -> bool:
    """Check if login is confirmed."""
    return LoginScopes.FRESH in connection.auth.scopes



import typing

from starlette.authentication import BaseUser
from starlette.datastructures import URL
from starlette.requests import Request
from starlette.responses import RedirectResponse
from starlette.types import ASGIApp, Receive, Scope, Send


class LoginRequiredMiddleware:
    def __init__(
        self,
        app: ASGIApp,
        redirect_url: str | None = None,
        *,
        path_name: str | None = "login",
        path_params: dict[str, typing.Any] | None = None,
    ) -> None:
        assert redirect_url or path_name
        self.app = app
        self.redirect_url = redirect_url
        self.path_name = path_name
        self.path_params = path_params or {}

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] not in {"http", "websocket"}:
            await self.app(scope, receive, send)
            return

        user = typing.cast(BaseUser, scope.get("user"))
        if not user.is_authenticated:
            request = Request(scope, receive, send)
            redirect_to = self.redirect_url or request.app.url_path_for(self.path_name, **self.path_params)
            url = URL(redirect_to).include_query_params(next=str(request.url.replace(scheme="", netloc="")))
            response = RedirectResponse(url, 302)
            await response(scope, receive, send)
        else:
            await self.app(scope, receive, send)
